use std::sync::Arc;

use cozo::DbInstance;
use tracing::debug;

use ubisync_lib::{
    api::events::AppEvent,
    peer::Peer,
    types::{Element, ElementContent, ElementId, MessageId, PeerId, PotId},
};

use crate::{node_events::UbisyncNodeEvent, state::queries};

use super::State;

pub struct CommState {
    state: Arc<State>,
}

impl CommState {
    pub fn new(state: Arc<State>) -> Self {
        CommState { state: state }
    }

    pub fn add_received_element(
        &self,
        id: &ElementId,
        content: &ElementContent,
        latest_message: &MessageId,
        pot_id: &PotId,
    ) -> anyhow::Result<()> {
        queries::elements::add(
            self.db(),
            &id,
            &content,
            Some(latest_message.to_owned()),
            false,
            pot_id,
        )?;
        debug!("Added element {{{}}}", &id.to_string());

        Ok(())
    }

    pub fn update_element_content(
        &self,
        id: &ElementId,
        content: &ElementContent,
        latest_message: &MessageId,
    ) -> anyhow::Result<()> {
        //TODO: resolve potential conflicts with local changes
        queries::elements::set_content(self.db(), id, content)?;
        queries::elements::set_latest_message(self.db(), id, Some(latest_message.to_owned()))?;
        debug!("Updated element {{{}}}", &id.to_string());

        let _ = self.state.emit_app_event(
            queries::pots::get_pot_members(
                self.db(),
                &queries::elements::get(self.db(), &id)?
                    .pot()
                    .as_ref()
                    .unwrap(),
            )?
            .get(0)
            .unwrap(),
            AppEvent::ElementUpdate { id: id.clone() },
        );

        Ok(())
    }

    pub fn remove_element(&self, id: &ElementId) -> anyhow::Result<()> {
        let res = self.state.remove_element(id);
        debug!("Removed element {{{}}}", &id.to_string());

        res
    }

    pub fn get_element(&self, id: &ElementId) -> anyhow::Result<Element> {
        self.state.get_element(id)
    }

    pub fn set_peer(&self, id: &PeerId, name: &str) -> anyhow::Result<()> {
        queries::peers::put(self.db(), id, name)?;
        debug!("Set peer {} with address {}.", &name, id.to_string());

        Ok(())
    }

    pub fn get_peers(&self) -> anyhow::Result<Vec<Peer>> {
        self.state.get_peers()
    }

    pub fn add_pot(&self, id: &PotId, app_type: &str) -> anyhow::Result<()> {
        queries::pots::add(self.db(), id, app_type)?;

        let _ = self.state.emit_node_event(UbisyncNodeEvent::NewPot {
            id: id.clone(),
            app_type: app_type.to_string(),
        });

        Ok(())
    }

    fn db(&self) -> &DbInstance {
        &self.state.db
    }
}

#[cfg(test)]
mod tests {
    use crate::state::State;

    use super::CommState;

    use tracing::Level;
    use ubisync_lib::types::{ElementContent, ElementId, MessageId, PotId};

    #[tokio::test]
    #[serial_test::serial]
    async fn test_element_add() {
        tracing_subscriber::fmt()
            .pretty()
            .with_max_level(Level::DEBUG)
            .init();

        let state = CommState::new(State::new("mem").await.unwrap());
        let id = ElementId::new();
        let pot_id = PotId::new();
        state
            .add_received_element(
                &id,
                &ElementContent::Text("Test-text".to_string()),
                &MessageId::new(),
                &pot_id,
            )
            .unwrap();
        let el = state.get_element(&id).unwrap();
        assert_eq!(
            ElementContent::Text("Test-text".to_string()),
            el.content().to_owned()
        )
    }

    #[tokio::test]
    #[serial_test::serial]
    async fn test_element_update() {
        tracing_subscriber::fmt()
            .pretty()
            .with_max_level(Level::DEBUG)
            .init();

        let state = CommState::new(State::new("mem").await.unwrap());
        let id = ElementId::new();
        let pot_id = PotId::new();
        state
            .add_received_element(
                &id,
                &ElementContent::Text("Test-text".to_string()),
                &MessageId::new(),
                &pot_id,
            )
            .unwrap();
        state
            .update_element_content(
                &id,
                &ElementContent::Text("Test-text 2".to_string()),
                &MessageId::new(),
            )
            .unwrap();
        let el = state.get_element(&id).unwrap();
        assert_eq!(
            ElementContent::Text("Test-text 2".to_string()),
            el.content().to_owned()
        )
    }
}
