use std::{
    collections::HashMap,
    sync::{
        mpsc::{channel, Receiver, Sender},
        Arc, Mutex, RwLock,
    },
};

use anyhow::Error;
use cozo::DbInstance;
use tracing::{debug, error, warn};

mod api_state;
mod comm_state;
mod queries;
mod schema;

pub use api_state::ApiState;
pub use comm_state::CommState;
use ubisync_lib::{
    api::events::AppEvent,
    messages::{Message, MessageContent},
    peer::Peer,
    types::{AppId, Element, ElementContent, ElementId, PotId, Tag},
};

use crate::{api::v0::app::App, comm::CommHandle, node_events::UbisyncNodeEvent, Ubisync};

pub struct State {
    db: DbInstance,
    comm_handle: RwLock<Option<Arc<CommHandle>>>,
    app_event_channels: RwLock<HashMap<AppId, (Sender<AppEvent>, Arc<Mutex<Receiver<AppEvent>>>)>>,
    node_event_callback: RwLock<Option<Box<dyn Fn(UbisyncNodeEvent) + Send + Sync>>>,
}

impl State {
    pub async fn new(db_location: &str) -> anyhow::Result<Arc<State>> {
        let db = match db_location {
            "mem" => DbInstance::new("mem", "", Default::default()),
            path => DbInstance::new("rocksdb", path, Default::default()),
        };
        match db {
            Ok(d) => {
                //TODO: change "add schema" to "ensure newest schema"
                schema::add_schema(&d)?;

                let state = Arc::new(State {
                    db: d,
                    comm_handle: RwLock::new(None),
                    app_event_channels: Default::default(),
                    node_event_callback: RwLock::new(None),
                });
                state.init_event_channels();
                Ok(state)
            }
            Err(e) => Err(Error::msg(format!("{:?}", e))),
        }
    }

    pub fn set_comm_handle(&self, handle: Arc<CommHandle>) {
        *self
            .comm_handle
            .write()
            .as_deref_mut()
            .expect("Could not set state's CommHandle") = Some(handle);
    }

    pub fn set_node_event_callback<T>(&self, cb: T, ubi: Arc<Ubisync>)
    where
        T: Fn(UbisyncNodeEvent, Arc<Ubisync>) + Send + Sync + 'static,
    {
        match self.node_event_callback.write() {
            Ok(mut writeguard) => {
                *writeguard = Some(Box::new(move |ev: UbisyncNodeEvent| (cb)(ev, ubi.clone())))
            }
            Err(e) => warn!("Failed to set node event callback: {:?}", e),
        }
    }

    pub fn get_apps(&self) -> anyhow::Result<Vec<App>> {
        queries::apps::get_all(&self.db)
    }

    pub fn set_element_content(
        &self,
        element_id: &ElementId,
        content: &ElementContent,
    ) -> anyhow::Result<()> {
        let res = queries::elements::set_content(&self.db, element_id, content);
        debug!(
            "Set content of element with id {:?}: {:?}",
            element_id,
            self.get_element(element_id)
        );

        self.send_to_peers(MessageContent::SetElement {
            id: element_id.clone(),
            content: content.clone(),
        });
        res
    }

    pub fn remove_element(&self, element_id: &ElementId) -> anyhow::Result<()> {
        let res = queries::elements::remove(&self.db, element_id);

        self.send_to_peers(MessageContent::RemoveElement {
            id: element_id.clone(),
        });
        res
    }

    pub fn get_element(&self, id: &ElementId) -> anyhow::Result<Element> {
        queries::elements::get(&self.db, id)
    }

    pub fn get_elements_by_tag(&self, tag: &Tag) -> Vec<ElementId> {
        queries::elements::get_by_tag(&self.db, tag)
            .map_err(|e| {
                error!("{}", e);
                e
            })
            .unwrap_or(vec![])
    }

    pub fn set_peer(&self, peer: &Peer) -> anyhow::Result<()> {
        queries::peers::put(&self.db, &peer.id(), &peer.name())
    }

    pub fn get_peers(&self) -> anyhow::Result<Vec<Peer>> {
        queries::peers::get(&self.db)
    }

    pub fn add_pot_member(&self, pot: &PotId, app: &AppId) -> anyhow::Result<()> {
        debug!("Pot: {:?}", queries::pots::get(&self.db, pot));
        match queries::pots::add_pot_member(&self.db, pot, app) {
            Ok(_) => {
                self.emit_app_event(
                    app,
                    AppEvent::NewPot {
                        id: pot.clone(),
                        app_type: queries::pots::get(&self.db, pot)?.app_type,
                    },
                );
                Ok(())
            }
            Err(e) => Err(e),
        }
    }

    pub fn get_event_receiver(&self, app: &AppId) -> anyhow::Result<Arc<Mutex<Receiver<AppEvent>>>> {
        self.create_event_channel_if_not_exists(app.clone());
        match self.app_event_channels.read() {
            Ok(map) => match map.get(&app) {
                Some((_, rx)) => Ok(rx.clone()),
                None => Err(Error::msg("Failed to get event receiver")),
            },
            Err(_) => Err(Error::msg("Failed to lock on event_channels")),
        }
    }

    pub fn emit_node_event(&self, ev: UbisyncNodeEvent) -> anyhow::Result<()> {
        debug!("1");
        match self.node_event_callback.read() {
            Ok(readguard) => {
                debug!("2");
                debug!("{:?}", readguard.is_some());
                // If a callback is set, call it.
                if let Some(cb) = readguard.as_ref() {
                    debug!("3");
                    (*cb)(ev);
                }
                // Whether a callback is set or not, return Ok(_)
                Ok(())
            }
            Err(_) => Err(Error::msg("Failed to lock on node event callback function")),
        }
    }

    fn emit_app_event(&self, app: &AppId, ev: AppEvent) {
        if let Ok(channels) = self.app_event_channels.read() {
            if let Some((sender, _)) = channels.get(app) {
                debug!("Emitting app event to {:?}", &app);
                if let Err(e) = sender.send(ev) {
                    debug!("Emitting app event failed: {:?}", e);
                }
            }
        }
        else {
            debug!("Failed to get channels");
        }
    }

    fn create_event_channel_if_not_exists(&self, app: AppId) {
        if let Ok(mut map) = self.app_event_channels.write() {
            if !map.contains_key(&app) {
                debug!(
                    "No event channel for {:?} exists, creating a new one.",
                    &app
                );
                let (tx, rx) = channel();
                map.insert(app, (tx, Arc::new(Mutex::new(rx))));
            }
        }
    }

    fn init_event_channels(&self) {
        let app_ids = queries::apps::get_all_ids(&self.db).unwrap_or(vec![]);
        for app in app_ids {
            self.create_event_channel_if_not_exists(app);
        }
    }

    fn send_to_peers(&self, ct: MessageContent) {
        match self.comm_handle.read() {
            Ok(opt) => {
                if opt.is_some() {
                    let arc = opt.as_ref().unwrap().clone();
                    tokio::spawn(async move {
                        let _ = arc.broadcast(Message::new(ct)).await;
                    });
                }
            }
            Err(e) => debug!("{}", e),
        }
    }
}
