use std::collections::BTreeMap;

use anyhow::{bail, Error};
use cozo::{DataValue, DbInstance, Num, ScriptMutability};
use itertools::Itertools;
use serde_with::chrono::{DateTime, Utc};
use tracing::{debug, warn};
use ubisync_lib::types::{AppId, Pot, PotId};

use crate::{api::v0::app::App, run_query};

pub fn add(
    db: &DbInstance,
    id: &AppId,
    last_access: &DateTime<Utc>,
    name: &str,
    description: &str,
    app_type: &str,
) -> anyhow::Result<()> {
    let params = vec![
        (
            "id",
            DataValue::Str(serde_json::to_string(&id).unwrap().into()),
        ),
        (
            "last_access",
            DataValue::Num(Num::Int(last_access.timestamp())),
        ),
        ("name", DataValue::Str(name.into())),
        ("description", DataValue::Str(description.into())),
        ("app_type", DataValue::Str(app_type.into())),
        ("default_pot", DataValue::Null),
    ];

    match run_query!(
        &db,
        ":insert apps {id => last_access, app_type, name, description, default_pot}",
        params,
        cozo::ScriptMutability::Mutable
    ) {
        Ok(_) => Ok(()),
        Err(report) => bail!(report),
    }
}

pub fn exists(db: &DbInstance, id: &AppId) -> anyhow::Result<bool> {
    let mut params = BTreeMap::new();
    params.insert(
        "id".to_string(),
        DataValue::Str(serde_json::to_string(&id)?.into()),
    );

    let result = db.run_script(
        "?[name] := *apps[$id, last_access, app_type, name, description, default_pot]",
        params,
        cozo::ScriptMutability::Immutable,
    );

    if let Ok(rows) = result {
        return Ok(rows.rows.len() == 1);
    }

    Err(Error::msg("Could not check whether app is registered"))
}

pub fn get_all(db: &DbInstance) -> anyhow::Result<Vec<App>> {
    match db.run_script(
        "?[id, last_access, app_type, name, description, default_pot] := *apps[id, last_access, app_type, name, description, default_pot]", 
        BTreeMap::default(), 
        ScriptMutability::Immutable) {
        Ok(named_rows) => {
            Ok(named_rows.rows.iter().filter_map(|row| {
                if let [
                    DataValue::Str(id), 
                    DataValue::Num(Num::Int(last_access)), 
                    DataValue::Str(app_type), 
                    DataValue::Str(name), 
                    DataValue::Str(description), 
                    default_pot,
                    ] = row.as_slice() {
                    Some(App {
                        id: serde_json::from_str(&id).unwrap(), 
                        app_type: app_type.to_string(), 
                        last_access: DateTime::from_timestamp(last_access.to_owned(), 0).unwrap(), 
                        name: name.to_string(), 
                        description: description.to_string(), 
                        default_pot: default_pot.get_str().map(|str| serde_json::from_str(str).unwrap())}
                    )
                }
                else {
                    None
                }
            }).collect_vec())
        },
        Err(e) => bail!(e),
    }
}

pub fn get_all_ids(db: &DbInstance) -> anyhow::Result<Vec<AppId>> {
    let result = db.run_script(
        "?[id] := *apps{id}",
        BTreeMap::new(),
        ScriptMutability::Immutable,
    );

    match result {
        Ok(named_rows) => Ok(named_rows
            .rows
            .iter()
            .filter_map(|row| {
                if let [DataValue::Str(app_id)] = row.as_slice() {
                    serde_json::from_str::<AppId>(app_id).ok()
                } else {
                    None
                }
            })
            .collect()),
        Err(e) => bail!(e),
    }
}

pub fn get(db: &DbInstance, id: &AppId) -> anyhow::Result<App> {
    let mut params = BTreeMap::new();
    params.insert(
        "id".to_string(),
        DataValue::Str(serde_json::to_string(&id)?.into()),
    );

    let result = db.run_script(
        "?[last_access, app_type, name, description, default_pot] := *apps[$id, last_access, app_type, name, description, default_pot]",
        params,
        cozo::ScriptMutability::Immutable,
    );

    if let Ok(rows) = result {
        if let Some(firstrow) = rows.rows.first() {
            if let [DataValue::Num(Num::Int(ts)), DataValue::Str(app_type), DataValue::Str(name), DataValue::Str(desc), default_pot] =
                firstrow.as_slice()
            {
                let last_access = DateTime::from_timestamp(ts.to_owned(), 0)
                    .ok_or(Error::msg("Failed to deserialize timestamp"))?;
                let pot_id = match default_pot {
                    DataValue::Str(dpid) => Some(serde_json::from_str(dpid)?),
                    _ => None,
                };
                return Ok(App {
                    id: id.clone(),
                    app_type: app_type.to_string(),
                    last_access,
                    name: name.to_string(),
                    description: desc.to_string(),
                    default_pot: pot_id,
                });
            }
        }
    }

    Err(Error::msg("Could not find app"))
}

pub fn set_default_pot(db: &DbInstance, id: &AppId, pot: &PotId) -> anyhow::Result<()> {
    let params = vec![
        ("id", DataValue::Str(serde_json::to_string(id)?.into())),
        (
            "default_pot",
            DataValue::Str(serde_json::to_string(pot)?.into()),
        ),
    ];

    match run_query!(
        &db,
        ":update apps {id => default_pot}",
        params.clone(),
        ScriptMutability::Mutable
    ) {
        Ok(_) => match run_query!(
            &db,
            ":put pot_memberships {pot_id = default_pot, app_id = id}",
            params,
            ScriptMutability::Mutable
        ) {
            Ok(_) => Ok(()),
            Err(report) => bail!(report),
        },
        Err(report) => bail!(report),
    }
}

pub fn get_default_pot(db: &DbInstance, app: &AppId) -> anyhow::Result<Pot> {
    let mut params = BTreeMap::new();
    params.insert(
        "app_id".to_string(),
        DataValue::Str(serde_json::to_string(&app)?.into()),
    );

    let result = db.run_script(
        "
        default_pot[pot_id] := *apps{id: $app_id, default_pot: pot_id}
        ?[pot_id, app_type] := default_pot[pot_id], *pots[pot_id, app_type]
    ",
        params.clone(),
        ScriptMutability::Immutable,
    );

    match result {
        Ok(rows) => {
            if let Some(firstrow) = rows.rows.first() {
                if let [DataValue::Str(pot_id_str), DataValue::Str(app_type)] = firstrow.as_slice()
                {
                    Ok(Pot {
                        id: serde_json::from_str(&pot_id_str)?,
                        app_type: app_type.to_string(),
                    })
                } else {
                    Err(Error::msg("Failed to deserialize query result"))
                }
            } else {
                Err(Error::msg("App not found"))
            }
        }
        Err(e) => bail!(e),
    }
}

pub fn create_pot(
    db: &DbInstance,
    pot_id: &PotId,
    app_id: &AppId,
    app_type: &str,
) -> anyhow::Result<()> {
    let params = vec![
        (
            "pot_id",
            DataValue::Str(serde_json::to_string(pot_id)?.into()),
        ),
        (
            "app_id",
            DataValue::Str(serde_json::to_string(app_id)?.into()),
        ),
        ("app_type", DataValue::Str(app_type.into())),
    ];

    match run_query!(
        &db,
        ":insert pots {id = pot_id => app_type}",
        params.clone(),
        ScriptMutability::Mutable
    ) {
        Ok(_) => match run_query!(
            &db,
            ":insert pot_memberships {pot_id => app_id}",
            params,
            ScriptMutability::Mutable
        ) {
            Ok(_) => Ok(()),
            Err(e) => {
                warn!("{:?}", e);
                bail!(e)
            }
        },
        Err(e) => {
            warn!("{:?}", e);
            bail!(e)
        }
    }
}

#[cfg(test)]
mod tests {
    use cozo::DbInstance;
    use itertools::Itertools;
    use serde_with::chrono::Utc;
    use tracing::{debug, Level};
    use ubisync_lib::types::{AppId, PotId};

    use crate::state::{queries::pots, schema};

    #[test]
    pub fn default_pot() {
        tracing_subscriber::fmt()
            .pretty()
            .with_max_level(Level::DEBUG)
            .init();

        let db = DbInstance::new("mem", "", Default::default()).unwrap();
        schema::add_schema(&db).unwrap();
        let app_id = AppId::new();
        let pot_id = PotId::new();
        super::add(&db, &app_id, &Utc::now(), "name", "description", "app_type").unwrap();
        pots::add(&db, &pot_id, "app_type").unwrap();
        super::set_default_pot(&db, &app_id, &pot_id).unwrap();

        debug!("Result: {:?}", super::get_default_pot(&db, &app_id));
    }

    #[test]
    pub fn add_and_get_all() {
        tracing_subscriber::fmt()
            .pretty()
            .with_max_level(Level::DEBUG)
            .init();

        let db = DbInstance::new("mem", "", Default::default()).unwrap();
        schema::add_schema(&db).unwrap();
        super::add(&db, &AppId::new(), &Utc::now(), "app1", "description", "app_type").unwrap();
        super::add(&db, &AppId::new(), &Utc::now(), "app2", "description", "app_type").unwrap();
        let all_ids = super::get_all_ids(&db).unwrap();
        let all_apps = super::get_all(&db).unwrap();
        assert_eq!(all_ids.len(), 2);
        assert_eq!(all_ids, all_apps.iter().map(|app| app.id.clone()).collect_vec());
    }
}
