use std::collections::BTreeMap;

use anyhow::Error;
use cozo::{DataValue, DbInstance, ScriptMutability};
use tracing::debug;
use ubisync_lib::types::{AppId, Pot, PotId};

use crate::run_query;

pub fn add(db: &DbInstance, id: &PotId, app_type: &str) -> anyhow::Result<()> {
    let params = vec![
        ("id", DataValue::Str(serde_json::to_string(id)?.into())),
        (
            "app_type",
            DataValue::Str(serde_json::to_string(app_type)?.into()),
        ),
    ];

    match run_query!(
        &db,
        ":put pots {id => app_type}",
        params,
        ScriptMutability::Mutable
    ) {
        Ok(_) => Ok(()),
        Err(report) => Err(Error::msg(format!("Query failed: {}", report))),
    }
}

pub fn get(db: &DbInstance, id: &PotId) -> anyhow::Result<Pot> {
    let mut params = BTreeMap::new();
    params.insert(
        "id".to_string(),
        DataValue::Str(serde_json::to_string(&id)?.into()),
    );
    let result = db.run_script(
        "
    ?[app_type] := *pots[$id, app_type]
    ",
        params,
        cozo::ScriptMutability::Immutable,
    );
    match result {
        Ok(rows) => {
            if let Some(firstrow) = rows.rows.first() {
                if let [DataValue::Str(app_type)] = firstrow.as_slice() {
                    Ok(Pot::new(
                        id.clone(),
                        serde_json::from_str(app_type)?,
                    ))
                } else {
                    Err(Error::msg("Could not parse result from query"))
                }
            } else {
                Err(Error::msg("Pot not found"))
            }
        }
        Err(report) => Err(Error::msg(format!("Query failed: {}", report))),
    }
}

pub fn add_pot_member(db: &DbInstance, pot: &PotId, app: &AppId) -> anyhow::Result<()> {
    let params = vec![
        ("pot_id", DataValue::Str(serde_json::to_string(pot)?.into())),
        ("app_id", DataValue::Str(serde_json::to_string(app)?.into())),
    ];

    match run_query!(
        &db,
        ":insert pot_memberships {pot_id, app_id}",
        params,
        ScriptMutability::Mutable
    ) {
        Ok(_) => Ok(()),
        Err(report) => Err(Error::msg(format!("Query failed: {}", report))),
    }
}

pub fn get_pot_members(db: &DbInstance, id: &PotId) -> anyhow::Result<Vec<AppId>> {
    let mut params = BTreeMap::new();
    params.insert(
        "id".to_string(),
        DataValue::Str(serde_json::to_string(&id)?.into()),
    );
    let result = db.run_script(
        "
    ?[app_id] := *pot_memberships{id: $id, app_id}
    ",
        params,
        cozo::ScriptMutability::Immutable,
    );
    match result {
        Ok(rows) => Ok(rows
            .rows
            .iter()
            .map(|row| {
                if let [DataValue::Str(app_id_string)] = row.as_slice() {
                    if let Ok(app_id) = serde_json::from_str::<AppId>(&app_id_string) {
                        Some(app_id)
                    } else {
                        None
                    }
                } else {
                    None
                }
            })
            .flatten()
            .collect()),
        Err(report) => Err(Error::msg(format!("Query failed: {}", report))),
    }
}

#[cfg(test)]
mod tests {
    use cozo::DbInstance;
    use ubisync_lib::types::PotId;

    use crate::state::schema;

    #[test]
    pub fn add_and_get() {
        let db = DbInstance::new("mem", "", Default::default()).unwrap();
        schema::add_schema(&db).unwrap();
        let pot_id = PotId::new();
        super::add(&db, &pot_id, "app_type").unwrap();
        let returned = super::get(&db, &pot_id).unwrap();
        assert_eq!(pot_id, returned.id);
        assert_eq!("app_type", returned.app_type);
    }
}